from __future__ import annotations

from abc import ABC, abstractmethod
from collections import deque
from collections.abc import AsyncIterator, Callable
from typing import ClassVar, Literal, TypeAlias, TypeVar

import anyio
from attrs import define, field
from sortedcontainers import SortedKeyList


@define
class Item(ABC):
    """
    The abstract class for items. Each subclass should represent one unit of work in
    your problem domain.
    """

    _set_worker_desc: Callable[[str], None] | None = field(default=None, init=False)
    _enqueue_fn: Callable[[Item], None] | None = field(default=None, init=False)

    @abstractmethod
    async def process(self) -> None:
        """
        Do this items work. This method is called when an item is popped from the queue.

        This method is required to be implemented.

        .. code-block:: python

            import aqueue

            class MyItem(aqueue.Item):
                async def process(self):
                    self.set_worker_desc("MyItem is processing...")

                    # do some work
                    ...

                    # enqueue more items
                    self.enqueue(Childitem())

            class ChildItem(aqueue.Item):
                async def process(self):
                    pass

        The return value for this method is aliased by `aqueue.ProcessRetVal`.
        """

    def enqueue(self, item: Item) -> None:
        """
        Enqueue another item for processing by a worker later. If this method is called
        outside of the process() method, it will raise a RuntimeError.

        :param item: The item to enqueue.
        """
        if self._enqueue_fn is None:
            raise RuntimeError("This function may only be called inside process()")
        self._enqueue_fn(item)

    def set_worker_desc(self, description: str) -> None:
        """
        Set the text description for the worker that is processing this item. If this
        method is called outside of the process() method, it will raise a RuntimeError.

        This method is a no-op if aqueue is not run with `run_queue`/`async_run_queue`
        is run with ``visualize=False``

        :param description: The text that the worker display will show.
        """
        if self._set_worker_desc is None:
            raise RuntimeError("This function may only be called inside process()")
        self._set_worker_desc(description)

    async def after_children_processed(self) -> None:
        """
        Do any kind of cleanup/finalization. This method is called after this item and
        all the child items enqueued by this item have returned from their ``process``
        method calls.

        The default implementation for this method is a no-op. Overriding it
        is optional.

        .. code-block:: python

            import aqueue

            class MyItem(aqueue.Item):
                async def after_children_processed(self):
                    print(f"{self} and all its children have been processed")
        """

    track_overall: ClassVar[bool] = False
    """
    Set this class variable to True to track it in the overall progress panel.

    If True, when this item is enqueued, the overall progress *total* will increment,
    and, when this item is done processing, the overall progress *completed* will
    increment.

    .. code-block:: python

        import aqueue

        class MyItem(aqueue.Item):
            track_overall = True
    """

    priority: ClassVar[int] = 0
    """
    Set this class variable to indicate priority.

    In priority queues, this number determines the ordering of how Item objects are
    popped from the queue for processing. Smaller numbers have higher priority.

    .. code-block:: python

        import aqueue

        class ImportantItem(aqueue.Item):
            priority = 1

        class PettyItem(aqueue.Item):
            priority = 2

    This attribute has no effect unless `run_queue`/`async_run_queue` is run with
    ``order="priority"``.
    """

    parent: Item | None = field(default=None, init=False)
    """
    The Item object that enqueued this one, or None if it was an initial item.

    This attribute is only valid inside ``process`` or after it has been called, such as
    in ``after_children_processed``. This attribute should not be overwritten or
    mutated.

    .. code-block:: python

        import aqueue

        class MyItem(aqueue.Item):
            async def process(self):
                parent = self.parent
                if not parent:
                    print("I'm an initial item")
                else:
                    print(f"I was created by {parent}")
    """

    _children: list[Item] = field(factory=list, init=False)
    """
    An internal list of Item objects that this item enqueued when it was processed.

    This is needed to keep track of children, which dictates when
    `after_children_processed` should be called.
    """

    _done_processing: bool = field(default=False, init=False)
    """
    An internal marker that is set to True when this method has completed processing.
    This is needed because, to properly trigger the `after_children_processed` callback,
    aqueue needs to know if an Item might still be processing.
    """

    @property
    def _tree_done(self) -> bool:
        """Return True if this Item and all its children have been processed."""
        return self._done_processing and all(
            child._tree_done for child in self._children
        )


# type for the return value of Item.process
ProcessRetVal: TypeAlias = AsyncIterator[Item] | None

QueueContainer = TypeVar("QueueContainer")


@define
class QueueABC(ABC):

    _unfinished_task_count: int = field(init=False, default=0)

    @abstractmethod
    def _put(self, item: Item) -> None:
        ...

    @abstractmethod
    def _get(self) -> Item:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    def put(self, item: Item) -> None:
        self._put(item)
        self._unfinished_task_count += 1

    async def get(self) -> Item:
        while self.empty():
            await anyio.sleep(0)
        return self._get()

    def empty(self) -> bool:
        return len(self) == 0

    def task_done(self) -> None:
        self._unfinished_task_count -= 1

    async def join(self) -> None:
        while self._unfinished_task_count != 0:
            await anyio.sleep(0)

    async def __aiter__(self) -> AsyncIterator[Item]:
        # gather 2 things: a get and a join
        not_changed = object()
        item: object | Item = not_changed

        async def join_and_cancel(cancel_scope: anyio.CancelScope) -> None:
            await self.join()
            cancel_scope.cancel()

        async def get_and_cancel(cancel_scope: anyio.CancelScope) -> None:
            nonlocal item
            item = await self.get()
            cancel_scope.cancel()

        while True:
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(join_and_cancel, task_group.cancel_scope)
                task_group.start_soon(get_and_cancel, task_group.cancel_scope)

            if item is not_changed:
                return
            else:
                yield item  # type: ignore

            item = not_changed  # reset for next iter


@define
class Queue(QueueABC):
    """
    An unbounded FIFO queue
    """

    _container: deque[Item] = field(factory=deque)

    def _put(self, item: Item) -> None:
        self._container.appendleft(item)

    def _get(self) -> Item:
        return self._container.pop()

    def __len__(self) -> int:
        return len(self._container)


@define
class Stack(QueueABC):
    """
    An unbounded LIFO queue (or stack)
    """

    _container: list[Item] = field(factory=list)

    def _put(self, item: Item) -> None:
        self._container.append(item)

    def _get(self) -> Item:
        return self._container.pop()

    def __len__(self) -> int:
        return len(self._container)


def _sortedkeylist_by_item() -> SortedKeyList:
    def key(value: Item) -> int:
        return value.priority

    return SortedKeyList(key=key)


@define
class PriorityQueue(QueueABC):
    """
    An unbounded priority queue. Items will need to set `priority` attributes
    appropriately.
    """

    _container: SortedKeyList = field(factory=_sortedkeylist_by_item)

    def _put(self, item: Item) -> None:
        self._container.add(item)

    def _get(self) -> Item:
        return self._container.pop(0)

    def __len__(self) -> int:
        return len(self._container)


# names a type of queue for the API
Ordering: TypeAlias = Literal["fifo", "lifo", "priority"]

QUEUE_FACTORY: dict[Ordering, type[QueueABC]] = {
    "fifo": Queue,
    "lifo": Stack,
    "priority": PriorityQueue,
}



from aqueue.processing import VisOptions, async_run_queue, run_queue
from aqueue.queue import Item, Ordering, ProcessRetVal

__all__ = (
    "ProcessRetVal",
    "Item",
    "Ordering",
    "run_queue",
    "async_run_queue",
    "VisOptions",
)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

import anyio
from attrs import define, frozen
from rich import get_console
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Table

if TYPE_CHECKING:
    # avoid circular reference when just trying to type hint
    from aqueue.queue import Item, QueueABC

WAIT_MESSAGE = "Waiting for work..."

# something obscure to not clash with user names
TOTAL_QUEUE_COUNT_NAME = "_aqueue_total"


@define(kw_only=True, slots=False)
class LinkedTask:
    """
    A progress and a task, linked together. This makes the rich API easier to work with
    because only one thing needs to be passed around, not two.
    """

    progress: Progress
    task_id: TaskID

    @classmethod
    def create(cls, progress: Progress, **kwargs: Any) -> LinkedTask:
        description = ""
        if "description" in kwargs:
            description = kwargs.pop("description")

        task_id = progress.add_task(description=description, **kwargs)

        return LinkedTask(progress=progress, task_id=task_id)

    @cached_property
    def task(self) -> Task:
        # yeesh, hacky. rich has questionable interface here
        return self.progress._tasks[self.task_id]

    def update(self, **kwargs: Any) -> None:
        self.progress.update(task_id=self.task_id, **kwargs)

    # convenience properties
    @property
    def completed(self) -> float:
        """Get the amount of completed items"""
        return self.task.completed

    @completed.setter
    def completed(self, value: float) -> None:
        """Set the amount of completed items"""
        self.update(completed=value)

    @property
    def total(self) -> float | None:
        """Get the amount of total items"""
        return self.task.total

    @total.setter
    def total(self, value: float | None) -> None:
        """Set the amount of total items"""
        self.update(total=value)

    @property
    def total_f(self) -> float:
        """Get the amount of total items. If the total is actually None, return 0."""
        return self.task.total or 0

    @total_f.setter
    def total_f(self, value: float) -> None:
        """Set the amount of total items. (Setter paired with total_f.)"""
        self.update(total=value)

    @property
    def description(self) -> str:
        """Get the current worker status description"""
        return self.task.description

    @description.setter
    def description(self, value: str) -> None:
        """Set the current worker status description"""
        self.update(description=value)


@define(kw_only=True)
class QueueItemCount:
    linked_task: LinkedTask

    @classmethod
    def create(cls, progress: Progress, name: str) -> QueueItemCount:
        linked_task = LinkedTask.create(progress, total=0, name=name)
        return QueueItemCount(linked_task=linked_task)

    @property
    def cur_count(self) -> float:
        return self.linked_task.completed

    @cur_count.setter
    def cur_count(self, value: int) -> None:
        self.linked_task.completed = value
        self.linked_task.total_f = max(value, self.linked_task.total_f)


@frozen(kw_only=True)
class Display:
    """Internal display object that organizes away the complex setup of it all."""

    queue: QueueABC
    live: Live
    overall_task: LinkedTask
    queue_stats_progress: Progress
    queue_stats_type_tasks: dict[str, QueueItemCount]
    worker_status_progress: Progress  # not a LinkedTask because workers wont yet exist
    console: Console

    @classmethod
    def create(cls, queue: QueueABC, console: Console | None = None) -> Display:
        if console is None:
            console = get_console()

        worker_status_progress = Progress(
            SpinnerColumn(),
            TextColumn("[blue]{task.fields[worker_id]}"),
            TextColumn("[white]{task.description}"),
            console=console,
        )

        queue_stats_progress = Progress(
            SpinnerColumn(),
            TextColumn("[blue]{task.fields[name]}"),
            TextColumn("[white]{task.completed} items cur"),
            BarColumn(complete_style="cyan", finished_style="cyan"),
            TextColumn("[white]{task.total} items max"),
            console=console,
        )
        queue_stats_type_tasks = {
            TOTAL_QUEUE_COUNT_NAME: QueueItemCount.create(
                queue_stats_progress, name="Total"
            )
        }

        overall_progress = Progress(
            SpinnerColumn(),
            TextColumn("[blue]{task.description}"),
            MofNCompleteColumn(),
            TaskProgressColumn(),
            BarColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=console,
        )
        overall_task = LinkedTask.create(progress=overall_progress, total=None)

        table = Table.grid()
        table.add_row(Panel(worker_status_progress, title="Worker Status"))
        table.add_row(Panel(queue_stats_progress, title="Queue Stats"))
        table.add_row(Panel(overall_progress, title="Overall Progress"))

        # with auto_refresh, we must call `refresh()`` manually or `update()`` with
        # `refresh=True`. But, this is fine because otherwise, rich starts a thread,
        # which is very anti-async.
        live = Live(table, auto_refresh=False, console=console)

        return Display(
            queue=queue,
            live=live,
            overall_task=overall_task,
            worker_status_progress=worker_status_progress,
            queue_stats_progress=queue_stats_progress,
            queue_stats_type_tasks=queue_stats_type_tasks,
            console=console,
        )

    def create_worker_status_task(
        self, worker_id: int, description: str = WAIT_MESSAGE
    ) -> LinkedTask:
        return LinkedTask.create(
            progress=self.worker_status_progress,
            description=description,
            worker_id=f"#{worker_id}",
        )

    def add_to_queue(self, item: Item) -> None:
        """Update the display to show that a new item has been added to the queue."""
        self.queue_stats_type_tasks[TOTAL_QUEUE_COUNT_NAME].cur_count += 1

        item_class_name = item.__class__.__qualname__
        if item_class_name not in self.queue_stats_type_tasks:
            self.queue_stats_type_tasks[item_class_name] = QueueItemCount.create(
                self.queue_stats_progress, item_class_name
            )

        self.queue_stats_type_tasks[item_class_name].cur_count += 1

    def remove_from_queue(self, item: Item) -> None:
        """Update the display to show that an item has been removed from the queue."""
        self.queue_stats_type_tasks[TOTAL_QUEUE_COUNT_NAME].cur_count -= 1
        self.queue_stats_type_tasks[item.__class__.__qualname__].cur_count -= 1

    def refresh(self) -> None:
        """Manually refresh the rich.live.Live object."""
        self.live.refresh()

    async def refresh_task(self, interval_sec: float = 1) -> None:
        now = anyio.current_time()
        while True:
            with anyio.CancelScope(shield=True):  # don't cancel during refresh
                self.refresh()
            await anyio.sleep_until(now + interval_sec)
            now = anyio.current_time()



from __future__ import annotations

from collections.abc import AsyncIterator, Iterable
from contextlib import asynccontextmanager
from functools import partial
from typing import Literal, TypeAlias

import anyio
from attrs import field, frozen
from rich.console import Console

from aqueue.display import WAIT_MESSAGE, Display, LinkedTask
from aqueue.queue import QUEUE_FACTORY, Item, Ordering, QueueABC


@frozen(kw_only=True)
class Worker:
    queue: QueueABC
    graceful_ctrl_c: bool

    async def loop_for_items(self) -> None:
        async for item in self.queue:
            self.before_item_process(item)
            item._set_worker_desc = self.set_worker_desc

            def enqueue(child: Item) -> None:
                child.parent = item
                self.queue.put(child)
                item._children.append(child)
                self.on_child_enqueued(child)

            item._enqueue_fn = enqueue

            with anyio.CancelScope(shield=self.graceful_ctrl_c):
                await item.process()
                item._done_processing = True

                cur_item: Item | None = item
                while cur_item:
                    if not cur_item._tree_done:
                        break
                    await cur_item.after_children_processed()
                    cur_item = cur_item.parent

            item._enqueue_fn = None
            item._set_worker_desc = None

            self.after_item_process(item)
            self.queue.task_done()

        self.on_worker_complete()

    def before_item_process(self, item: Item) -> None:
        pass

    def set_worker_desc(self, description: str) -> None:
        pass

    def on_child_enqueued(self, item: Item) -> None:
        pass

    def after_item_process(self, item: Item) -> None:
        pass

    def on_worker_complete(self) -> None:
        pass


@frozen(kw_only=True)
class VisWorker(Worker):
    display: Display
    worker_status_task: LinkedTask

    def set_worker_desc(self, description: str) -> None:
        self.worker_status_task.description = description

    def before_item_process(self, item: Item) -> None:
        self.display.remove_from_queue(item)

    def on_child_enqueued(self, item: Item) -> None:
        self.display.add_to_queue(item)
        if item.track_overall:
            self.display.overall_task.total_f += 1

    def after_item_process(self, item: Item) -> None:
        if item.track_overall:
            self.display.overall_task.completed += 1
        self.worker_status_task.description = WAIT_MESSAGE

    def on_worker_complete(self) -> None:
        self.worker_status_task.description = "Done"


@frozen(kw_only=True)
class QueueRunner:
    num_workers: int = 5
    queue: QueueABC
    graceful_ctrl_c: bool = True

    @classmethod
    def _init_queue(
        cls,
        initial_items: Iterable[Item],
        order: Ordering,
    ) -> QueueABC:
        queue = QUEUE_FACTORY[order]()
        for item in initial_items:
            queue.put(item)
        return queue

    @classmethod
    def create(
        cls,
        initial_items: Iterable[Item],
        num_workers: int,
        order: Ordering,
        graceful_ctrl_c: bool,
    ) -> QueueRunner:
        queue = cls._init_queue(initial_items=initial_items, order=order)
        return QueueRunner(
            num_workers=num_workers, queue=queue, graceful_ctrl_c=graceful_ctrl_c
        )

    @asynccontextmanager
    async def outside_worker_loop_context(self) -> AsyncIterator[None]:
        """A context manager that is entered/exited outside the main worker loop."""
        yield

    def create_worker(self, worker_id: int) -> Worker:
        return Worker(queue=self.queue, graceful_ctrl_c=self.graceful_ctrl_c)

    def after_loop(self) -> None:
        pass

    async def run(self) -> None:
        async with self.outside_worker_loop_context():
            async with anyio.create_task_group() as worker_task_group:
                for worker_id in range(self.num_workers):
                    worker = self.create_worker(worker_id)
                    worker_task_group.start_soon(
                        worker.loop_for_items,
                        name=f"worker #{worker_id}",  # for debugging/introspection
                    )
        self.after_loop()


@frozen(kw_only=True)
class VisOptions:
    """
    A container for settings when visualizing the queue.
    """

    console: Console | None = field(default=None)
    """
    An explicit `rich.console.Console` to print the visualization. This should be
    provided if your are using a Console elsewhere in your code because it will conflict
    with aqueue's internal one.

    Setting this to `None` will use the console returned by `rich.get_console`.
    """


@frozen(kw_only=True)
class VisQueueRunner(QueueRunner):
    display: Display

    @classmethod
    def create(
        cls,
        initial_items: Iterable[Item],
        num_workers: int,
        order: Ordering,
        graceful_ctrl_c: bool,
        vis_options: VisOptions | None = None,
    ) -> VisQueueRunner:
        queue = cls._init_queue(initial_items=initial_items, order=order)

        console = None
        if vis_options and vis_options.console:
            console = vis_options.console

        display = Display.create(queue=queue, console=console)
        display.live.start()

        for item in initial_items:
            display.add_to_queue(item)
            if item.track_overall:
                display.overall_task.total_f += 1

        return VisQueueRunner(
            num_workers=num_workers,
            queue=queue,
            graceful_ctrl_c=graceful_ctrl_c,
            display=display,
        )

    @asynccontextmanager
    async def outside_worker_loop_context(self) -> AsyncIterator[None]:
        async with anyio.create_task_group() as live_refresher_task_group:
            live_refresher_task_group.start_soon(
                self.display.refresh_task, name="live-refresher"
            )
            yield
            live_refresher_task_group.cancel_scope.cancel()

    def create_worker(self, worker_id: int) -> Worker:
        return VisWorker(
            queue=self.queue,
            graceful_ctrl_c=self.graceful_ctrl_c,
            display=self.display,
            worker_status_task=self.display.create_worker_status_task(
                worker_id=worker_id
            ),
        )

    def after_loop(self) -> None:
        self.display.live.stop()


async def async_run_queue(
    *,
    initial_items: Iterable[Item],
    num_workers: int = 5,
    order: Ordering = "lifo",
    graceful_ctrl_c: bool = True,
    visualize: bool | VisOptions = True,
) -> None:
    """
    The coroutine of `aqueue.run_queue`. This function is useful if you want to bring
    your own event loop — either `asyncio` or `trio` event loops are usable with
    ``aqueue`` (those provided by `AnyIO <https://anyio.readthedocs.io/>`_).

    The parameters for both functions are the same.

    .. code-block:: python

        import asyncio, aqueue

        class MyItem(aqueue.Item):
            async def process(self):
                pass

        asyncio.run(aqueue.async_run_queue(
            initial_items=[MyItem()],
        ))
    """
    if visualize:
        vis_opions = visualize
        if vis_opions is True:
            vis_opions = VisOptions()

        runner: QueueRunner = VisQueueRunner.create(
            initial_items=initial_items,
            num_workers=num_workers,
            order=order,
            graceful_ctrl_c=graceful_ctrl_c,
            vis_options=vis_opions,
        )
    else:
        runner = QueueRunner.create(
            initial_items=initial_items,
            num_workers=num_workers,
            order=order,
            graceful_ctrl_c=graceful_ctrl_c,
        )

    await runner.run()


AsyncBackend: TypeAlias = Literal["trio", "asyncio"]


def run_queue(
    *,
    initial_items: Iterable[Item],
    num_workers: int = 5,
    order: Ordering = "lifo",
    graceful_ctrl_c: bool = True,
    visualize: bool | VisOptions = True,
    async_backend: AsyncBackend = "asyncio",
) -> None:
    """
    Process all items in ``initial_items`` (and any subsequent items they enqueue) until
    they are complete. Meanwhile, display a terminal visualization of it if
    ``visualize`` is not `False`.

    .. code-block:: python

        import aqueue

        class MyItem(aqueue.Item):
            async def process(self):
                pass

        aqueue.run_queue(
            initial_items=[MyItem()],
        )

    :param initial_items: An iterable that seeds the queue. This is where the top-level
        item(s) should go that produces more items.

    :param num_workers: Specifies how many workers will be running concurrently.

        .. note::

           Setting a high ``num_workers`` does not automatically mean your program will
           run faster.

    :param order: Can be either of:

        - ``lifo`` for last-in-first-out processing, or depth-first. This is recommended
          for website scraping because it yields your deepest items fast, which are
          probably what you're really after (versus ``fifo`` that processes intermediate
          items first). This is the default.

        - ``fifo`` for first-in-first-out processing, or breadth-first.

        - ``priority`` for priority-based processing. In this case, processing will
          occur by *ascending* `Item.priority` (smallest first).

        .. note::

            ``priority`` only guarantees that workers will select items from the queue
            according to priority. It does not guarantee that items will be completed in
            any order.

        The type for this argument is aliased by `aqueue.Ordering`.

    :param graceful_ctrl_c: specifies whether pressing Ctrl-C will stop things abruptly
        (`False`) or wait until all workers' current items have finished their process()
        methods and their after_children_processed() methods (`True`, the default). The
        latter will give you more opportunity to clean up resources, write data, etc.

        .. warning::

           If you write a buggy item that never finishes, Ctrl-C will have no effect.

    :param visualize: Set to `True` to draw a live progress meter display in the
        terminal while the queue is processed. Setting this to `False` does none of
        that.

        Further, you may pass an `aqueue.VisOptions` object here, which acts just like
        `True`, but lets you configure some things about how the visualization is
        performed.

    :param async_backend: Specified the asynchronous backend framework with which to run
        the event loop. Generally speaking, you may not mix-and-match event loops and
        the async primitives run inside them. Therefore, this argument lets you choose
        which one to use.

        The supported backends are those provided by `AnyIO`_.

    """

    anyio.run(
        partial(
            async_run_queue,
            order=order,
            initial_items=initial_items,
            num_workers=num_workers,
            graceful_ctrl_c=graceful_ctrl_c,
            visualize=visualize,
        ),
        backend=async_backend,
    )
